package lib

import (
	"bytes"
	"fmt"

	"github.com/Masterminds/semver"
)

func BuildDependencyTree(rp PackageJSON) (*DependencyList, error) {
	resolved := map[string]Dependency{}
	unresolved := []Dependency{
		Dependency{Name: rp.Name, Requires: rp.Dependencies},
	}

	for len(unresolved) > 0 {
		unresolvedDep := unresolved[0]
		unresolved = unresolved[1:len(unresolved)]

		for k, p := range unresolvedDep.Requires {
			meta, err := GetPackageInfo(k)
			if err != nil {
				return nil, err
			}

			c, err := semver.NewConstraint(p)
			if err != nil {
				return nil, err
			}

			var selectedRelease PackageRelease
			var max *semver.Version
			for r, release := range meta.Versions {
				v, err := semver.NewVersion(r)
				if err != nil {
					return nil, err
				}

				if c.Check(v) && (max == nil || v.GreaterThan(max)) {
					selectedRelease = release
					max = v
				}
			}

			unresolved = append(unresolved, Dependency{
				Name:      selectedRelease.Name,
				Version:   selectedRelease.VersionNumber,
				Integrity: selectedRelease.Dist.Sum,
				Resolved:  selectedRelease.Dist.Tarball,
				Requires:  selectedRelease.Dependencies,
			})
		}

		// don't add root dep to the resolve list
		if unresolvedDep.Name == rp.Name {
			continue
		}

		resolved[unresolvedDep.Name] = unresolvedDep
	}

	return &DependencyList{
		Name:            rp.Name,
		Version:         rp.Version,
		LockfileVersion: 1,
		Requires:        true,
		Dependencies:    resolved,
	}, nil
}

type Dependency struct {
	Name      string
	Version   string
	Integrity string
	Resolved  string
	Requires  RequireList
}

func (d Dependency) TarURL() string {
	return fmt.Sprintf("https://registry.npmjs.org/%s/-/%[1]s-%s.tgz", d.Name, d.Version)
}

func (d Dependency) String() string {
	header := fmt.Sprintf("%s@%s - %s\n", d.Name, d.Version, d.Integrity)
	buf := bytes.NewBufferString(header)

	for k, v := range d.Requires {
		fmt.Fprintf(buf, "\t%s => %s\n", k, v)
	}

	return buf.String()
}

type RequireList map[string]string

type DependencyList struct {
	Name            string                `json:"name"`
	Version         string                `json:"version"`
	LockfileVersion int                   `json:"lockfileVersion"`
	Requires        bool                  `json:"requires"`
	Dependencies    map[string]Dependency `json:"dependencies"`
}
